from collections import defaultdict
from math import sqrt
from typing import Dict, List, Optional

query = "TRUNCATE TABLE algoritmos_arbol.estadisticas_evaluacion"
plpy.execute(query)


class ResultadoEstadistico:
    vp: int
    fp: int
    vn: int
    fn: int

    def __init__(self, vp: int, fp: int, vn: int, fn: int):
        self.vp = vp
        self.fp = fp
        self.vn = vn
        self.fn = fn

    @classmethod
    def desde_clase(cls, clase_buscada: str, matriz: Dict[str, Dict[str, int]]):
        vp = 0
        fp = 0
        vn = 0
        fn = 0

        for clase_fila, columnas in matriz.items():
            if clase_fila == clase_buscada:
                for columna, cantidad in columnas.items():

                    if columna == clase_buscada:
                        vp += cantidad
                    else:
                        fn += cantidad
            else:
                for columna, cantidad in columnas.items():
                    if columna == clase_buscada:
                        fp += cantidad
                    else:
                        vn += cantidad

        return cls(vp, fp, vn, fn)

    def total_clase(self):
        return self.vp + self.fn

    def tpr(self) -> Optional[float]:
        vp = self.vp
        fn = self.fn
        numerador = vp
        denominador = vp + fn
        if denominador == 0:
            return None
        return numerador / denominador

    def recall(self) -> Optional[float]:
        return self.tpr()  # Recall es igual a TPR

    def tnr(self) -> Optional[float]:
        fp = self.fp
        vn = self.vn
        numerador = vn
        denominador = vn + fp
        if denominador == 0:
            return None
        return numerador / denominador

    def fpr(self) -> Optional[float]:
        fp = self.fp
        vn = self.vn
        numerador = fp
        denominador = fp + vn
        if denominador == 0:
            return None
        return numerador / denominador

    def fnr(self) -> Optional[float]:
        vp = self.vp
        fn = self.fn
        numerador = fn
        denominador = fn + vp
        if denominador == 0:
            return None
        return numerador / denominador

    def presicion(self) -> Optional[float]:
        vp = self.vp
        fp = self.fp
        numerador = vp
        denominador = vp + fp
        if denominador == 0:
            return None
        return numerador / denominador

    def f_score(self) -> Optional[float]:
        recall = self.tpr()
        presicion = self.presicion()
        if recall is None or presicion is None:
            return None
        numerador = 2 * recall * presicion
        denominador = recall + presicion
        if denominador == 0:
            return None
        return numerador / denominador

    def mcc(self) -> Optional[float]:
        vp = self.vp
        fp = self.fp
        vn = self.vn
        fn = self.fn
        numerador = (vp * vn) - (fp * fn)
        denominador = sqrt((vp + fp) * (vp + fn) * (vn + fp) * (vn + fn))
        if denominador == 0:
            return None
        return numerador / denominador

    def __repr__(self) -> str:
        return f"<vp={self.vp} fp={self.fp} vn={self.vn} fn={self.fn}>"


class Prediccion:
    clase_actual: str
    probabilidades: Dict[str, float]
    cantidad: int

    predicciones: List["Prediccion"] = []

    def __init__(
        self,
        clase_actual: str,
        probabilidades: Dict[str, float],
        cantidad: int,
    ) -> None:
        self.clase_actual = clase_actual
        self.probabilidades = probabilidades
        self.cantidad = cantidad

    def __repr__(self) -> str:
        return f"<clase_actual={self.clase_actual} cantidad={self.cantidad} probs={self.probabilidades}>"

    @staticmethod
    def adicionar_prediccion(
        clase_actual: str, probabilidades: Dict[str, float], cantidad: int
    ):
        prediccion = Prediccion(clase_actual, probabilidades, cantidad)
        Prediccion.predicciones.append(prediccion)

    @staticmethod
    def obtener_ordenado(clase: str) -> List["Prediccion"]:
        lista_ordenada = sorted(
            Prediccion.predicciones, key=lambda pred: pred.probabilidades[clase]
        )
        return lista_ordenada


class Instancia:
    verdaderos_positivos: int
    falsos_positivos: int
    recall: Optional[float]
    presicion: Optional[float]

    def __init__(
        self,
        positivos_verdaderos: int,
        positivos_falsos: int,
        recall: Optional[float],
        presicion: Optional[float],
    ) -> None:
        self.verdaderos_positivos = positivos_verdaderos
        self.falsos_positivos = positivos_falsos
        self.recall = recall
        self.presicion = presicion

    def __repr__(self) -> str:
        recall = f"{self.recall:.3f}" if self.recall is not None else "?"
        presicion = f"{self.presicion:.3f}" if self.presicion is not None else "?"
        return f"<tp={self.verdaderos_positivos} fp={self.falsos_positivos} recall={recall} presicion={presicion}>"


class UmbralCurva:
    predicciones: List[Prediccion]
    instancias: List[Instancia]
    clase: str

    def __init__(self, clase: str) -> None:
        self.predicciones = Prediccion.obtener_ordenado(clase)
        self.clase = clase
        self.crear_curva()

    def crear_curva(self) -> None:
        self.instancias = []

        # Obtener total positivos y total negativos
        query = f"""
            SELECT clase, cantidad
            FROM algoritmos_arbol.cantidades_anteriores
        """
        resultado = plpy.execute(query)

        total_positivos = 0
        total_negativos = 0

        for fila in resultado:
            if fila["clase"] == self.clase:
                total_positivos += fila["cantidad"]
            else:
                total_negativos += fila["cantidad"]

        tc = ResultadoEstadistico(total_positivos, total_negativos, 0, 0)

        positivos = 0
        negativos = 0

        i = 0

        while True:
            tc.vp = tc.vp - positivos
            tc.fn = tc.fn + positivos
            tc.fp = tc.fp - negativos
            tc.vn = tc.vn + negativos

            instancia = Instancia(tc.vp, tc.fp, tc.recall(), tc.presicion())
            self.instancias.append(instancia)

            positivos = 0
            negativos = 0

            iguales = True
            while iguales:
                prediccion = self.predicciones[i]
                if prediccion.clase_actual == self.clase:
                    positivos += prediccion.cantidad
                else:
                    negativos += prediccion.cantidad

                if i == len(self.predicciones) - 1 or (
                    self.predicciones[i + 1].probabilidades[clase]
                    != prediccion.probabilidades[clase]
                ):
                    iguales = False

                i += 1

            if i == len(self.predicciones):
                break

        if tc.fn != total_positivos or tc.vn != total_negativos:
            tc = ResultadoEstadistico(0, 0, total_negativos, total_positivos)
            instancia = Instancia(tc.vp, tc.fp, tc.recall(), tc.presicion())
            self.instancias.append(instancia)

    def area_roc(self) -> Optional[float]:
        n = len(self.instancias)

        if n == 0:
            return None

        area = 0
        cum_neg = 0

        instancias = self.instancias

        total_pos = instancias[0].verdaderos_positivos
        total_neg = instancias[0].falsos_positivos

        for i in range(n):
            if i < n - 1:
                cip = (
                    instancias[i].verdaderos_positivos
                    - instancias[i + 1].verdaderos_positivos
                )
                cin = (
                    instancias[i].falsos_positivos - instancias[i + 1].falsos_positivos
                )
            else:
                cip = instancias[n - 1].verdaderos_positivos
                cin = instancias[n - 1].falsos_positivos

            area += cip * (cum_neg + (0.5 * cin))
            cum_neg += cin

        if total_neg * total_pos == 0:
            return None

        area /= total_neg * total_pos

        return area

    def area_prc(self) -> Optional[float]:
        n = len(self.instancias)

        if n == 0:
            return None

        area = 0

        instancias = self.instancias

        xlast = instancias[n - 1].recall

        for i in range(n - 2, -1, -1):
            recall = instancias[i].recall
            if recall is None or xlast is None:
                return None

            recall_delta = recall - xlast

            presicion = instancias[i].presicion
            if presicion is None or xlast is None:
                return None

            area += presicion * recall_delta

            xlast = instancias[i].recall

        if area == 0:
            return None

        return area

    def __repr__(self) -> str:
        texto = f"predicciones: {self.predicciones}\n"
        texto += f"instancias: {self.instancias}\n"
        texto += f"clase: {self.clase}\n"

        return texto


class Estadisticas:
    suma_errores: float
    suma_errores_abs: float
    suma_errores_sqr: float
    suma_errores_anterior_abs: float
    suma_errores_anterior_sqr: float

    total_matriz: int
    total_tabla: int

    total_correctas: int
    total_incorrectas: int

    def __init__(self, matriz: Dict[str, Dict[str, int]]) -> None:
        self.actualizar_errores(matriz)

    def actualizar_errores(self, matriz: Dict[str, Dict[str, int]]):
        self.suma_errores = 0
        self.suma_errores_abs = 0
        self.suma_errores_sqr = 0
        self.suma_errores_anterior_abs = 0
        self.suma_errores_anterior_sqr = 0

        total_correctas = 0
        total_incorrectas = 0

        for fila, dict_col in matriz.items():
            for columna, valor in dict_col.items():
                if fila == columna:
                    total_correctas += valor
                else:
                    total_incorrectas += valor

        self.total_correctas = total_correctas
        self.total_incorrectas = total_incorrectas

        query = f"""
            SELECT clase, cantidad
            FROM algoritmos_arbol.cantidades_anteriores
        """

        resultado = plpy.execute(query)
        total: int = sum(fila["cantidad"] for fila in resultado)

        probabilidades_anteriores: dict[str, float] = defaultdict(float)
        for fila in resultado:
            probabilidades_anteriores[fila["clase"]] = fila["cantidad"] / total

        self.total_matriz = total_correctas + total_incorrectas
        self.total_tabla = total

        for prediccion in Prediccion.predicciones:
            suma_errores: float = 0
            suma_errores_abs: float = 0
            suma_errores_sqr: float = 0
            suma_errores_anterior_abs: float = 0
            suma_errores_anterior_sqr: float = 0

            for clase, probabilidad in prediccion.probabilidades.items():
                if clase == prediccion.clase_actual:
                    probabilidad_actual = 1
                else:
                    probabilidad_actual = 0

                diferencia = probabilidad - probabilidad_actual
                suma_errores += diferencia
                suma_errores_abs += abs(diferencia)
                suma_errores_sqr += diferencia**2

                diferencia = probabilidades_anteriores[clase] - probabilidad_actual
                suma_errores_anterior_abs += abs(diferencia)
                suma_errores_anterior_sqr += diferencia**2

            cantidad_probs = len(prediccion.probabilidades)

            self.suma_errores += suma_errores * prediccion.cantidad / cantidad_probs
            self.suma_errores_abs += (
                suma_errores_abs * prediccion.cantidad / cantidad_probs
            )
            self.suma_errores_sqr += (
                suma_errores_sqr * prediccion.cantidad / cantidad_probs
            )
            self.suma_errores_anterior_abs += (
                suma_errores_anterior_abs * prediccion.cantidad / cantidad_probs
            )
            self.suma_errores_anterior_sqr += (
                suma_errores_anterior_sqr * prediccion.cantidad / cantidad_probs
            )

    @staticmethod
    def kappa(matriz: Dict[str, Dict[str, int]]) -> float:
        etiquetas = list(matriz.keys())

        suma_filas = {clase: 0 for clase in etiquetas}
        suma_columnas = {clase: 0 for clase in etiquetas}
        suma_pesos = 0

        for fila in etiquetas:
            for columna in etiquetas:
                suma_filas[fila] += matriz[fila][columna]
                suma_columnas[columna] += matriz[fila][columna]
                suma_pesos += matriz[fila][columna]

        correcto = 0
        acuerdo_azar = 0

        for etiqueta in etiquetas:
            acuerdo_azar += suma_filas[etiqueta] * suma_columnas[etiqueta]
            correcto += matriz[etiqueta][etiqueta]

        acuerdo_azar /= suma_pesos * suma_pesos
        correcto /= suma_pesos

        if acuerdo_azar < 1:
            return (correcto - acuerdo_azar) / (1 - acuerdo_azar)
        else:
            return 1

    def error_absoluto_medio(self) -> float:
        return self.suma_errores_abs / self.total_matriz

    def error_absoluto_medio_anterior(self) -> float:
        return self.suma_errores_anterior_abs / self.total_tabla

    def raiz_error_absoluto_medio_cuadratico(self) -> float:
        return sqrt(self.suma_errores_sqr / self.total_matriz)

    def raiz_error_absoluto_medio_cuadratico_anterior(self) -> float:
        return sqrt(self.suma_errores_anterior_sqr / self.total_tabla)

    def error_absoluto_relativo(self) -> float:
        error = self.error_absoluto_medio_anterior()
        if error == 0:
            return 0
        return 100 * self.error_absoluto_medio() / self.error_absoluto_medio_anterior()

    def raiz_error_cuadratico_relativo(self) -> float:
        error = self.raiz_error_absoluto_medio_cuadratico_anterior()
        if error == 0:
            return 0

        return 100 * self.raiz_error_absoluto_medio_cuadratico() / error


# ========== Obtener matriz de confusion =========

query = """
    SELECT fila, columna, cantidad
    FROM algoritmos_arbol.matriz_confusion
"""

resultado = plpy.execute(query)

matriz: Dict[str, Dict[str, int]] = defaultdict(dict)
total_clases = 0

for fila_resultado in resultado:
    fila: str = fila_resultado["fila"]
    columna: str = fila_resultado["columna"]
    cantidad: int = fila_resultado["cantidad"]

    matriz[fila][columna] = cantidad
    total_clases += cantidad

# ========== Obtener Predicciones ===========
query = """
    SELECT id, clase, cantidad
    FROM algoritmos_arbol.clasificados
"""
resultado = plpy.execute(query)

for fila_resultado in resultado:
    query = f"""
        SELECT clase::text, probabilidad
        FROM algoritmos_arbol.clasificados_probabilidades
        WHERE id = {fila_resultado["id"]}
    """
    resultado_probabilidades = plpy.execute(query)

    probabilidades: Dict[str, float] = {
        fila["clase"]: fila["probabilidad"] for fila in resultado_probabilidades
    }

    Prediccion.adicionar_prediccion(
        fila_resultado["clase"], probabilidades, fila_resultado["cantidad"]
    )


# ======== Hacer los demas calculos =======


def actualizar_total_ponderado(
    total: Optional[float], valor: Optional[float], porcentaje: float
) -> Optional[float]:
    return (
        total + valor * porcentaje
        if isinstance(total, float) and isinstance(valor, float)
        else None
    )


totales_ponderados: Dict[str, float | None] = {
    "TPR": 0.0,
    "FPR": 0.0,
    "Precision": 0.0,
    "Recall": 0.0,
    "F-Measure": 0.0,
    "MCC": 0.0,
    "ROC-Area": 0.0,
    "PRC-Area": 0.0,
}


def verificar_entrada(valor: float | None):
    if valor is None:
        return "NULL"
    else:
        return valor


def insertar_en_tabla(diccionario, tipo: str, clase: str):
    query = """
        INSERT INTO algoritmos_arbol.estadisticas_evaluacion
        (
            tipo,
            clase,
            tpr,
            fpr,
            presicion,
            recall,
            f_measure,
            mcc,
            roc_area,
            prc_area
        )
        VALUES ('%s', '%s', %s, %s, %s, %s, %s, %s, %s, %s)
    """ % (
        tipo,
        clase,
        verificar_entrada(diccionario["TPR"]),
        verificar_entrada(diccionario["FPR"]),
        verificar_entrada(diccionario["Precision"]),
        verificar_entrada(diccionario["Recall"]),
        verificar_entrada(diccionario["F-Measure"]),
        verificar_entrada(diccionario["MCC"]),
        verificar_entrada(diccionario["ROC-Area"]),
        verificar_entrada(diccionario["PRC-Area"]),
    )

    plpy.execute(query)


for clase in matriz.keys():
    resultado = ResultadoEstadistico.desde_clase(clase, matriz)
    porcentaje = resultado.total_clase() / total_clases

    curva = UmbralCurva(clase)

    metricas = {
        "TPR": resultado.tpr(),
        "FPR": resultado.fpr(),
        "Precision": resultado.presicion(),
        "Recall": resultado.recall(),
        "F-Measure": resultado.f_score(),
        "MCC": resultado.mcc(),
        "ROC-Area": curva.area_roc(),
        "PRC-Area": curva.area_prc(),
    }

    for nombre, valor in metricas.items():
        totales_ponderados[nombre] = actualizar_total_ponderado(
            totales_ponderados[nombre], valor, porcentaje
        )

    # Añadir a base de datos
    insertar_en_tabla(metricas, "PARCIAL", clase)


insertar_en_tabla(totales_ponderados, "TOTAL", "Total Ponderado")


# ========== Estadisticas Generales ==========

# Eliminar estadisticas anteriores
plpy.execute("TRUNCATE TABLE algoritmos_arbol.info_evaluacion")


def insertar_info(id: str, titulo: str, valor: float):
    plpy.execute(
        """
        INSERT INTO algoritmos_arbol.info_evaluacion(id, titulo, valor)
        VALUES ('%s', '%s', %s)
    """
        % (id, titulo, valor)
    )


estadisticas = Estadisticas(matriz)
insertar_info(
    "inst_correctas",
    "Instancias Clasificadas Correctamente",
    estadisticas.total_correctas,
)
insertar_info(
    "inst_incorrectas",
    "Instancias Clasificadas Incorrectamente",
    estadisticas.total_incorrectas,
)
insertar_info("kappa", "Estadistica Kappa", estadisticas.kappa(matriz))
insertar_info(
    "err_abs_medio", "Error Absoluto Medio", estadisticas.error_absoluto_medio()
)
insertar_info(
    "raiz_err_abs_cuad",
    "Raiz del Error Medio Cuadratico",
    estadisticas.raiz_error_absoluto_medio_cuadratico(),
)
insertar_info(
    "err_abs_rel", "Error Absoluto Relativo", estadisticas.error_absoluto_relativo()
)
insertar_info(
    "raiz_err_cuad_rel",
    "Raiz del Error Cuadratico Relativo",
    estadisticas.raiz_error_cuadratico_relativo(),
)
insertar_info("inst_total", "Total de Instancias", estadisticas.total_matriz)
