# tabla = "test"
# atributo_objetivo = "test2"
# esquema = "public"

from collections import defaultdict
from typing import Dict, List, Literal, Optional, Tuple
from math import log2

# Eliminar arbol anterior
query = "TRUNCATE algoritmos_arbol.nodos;"
plpy.execute(query)
query = "TRUNCATE algoritmos_arbol.info;"
plpy.execute(query)

# Obtener clases posibles
query = f"""
    SELECT DISTINCT {atributo_objetivo}::text as clase
    FROM {esquema}.{tabla}
"""
resultado = plpy.execute(query)

clases_unicas: Dict[str, float] = {fila["clase"]: 0 for fila in resultado}


class Probabilidad:
    probabilidades: Dict[int, Dict[str, float]] = defaultdict(lambda: clases_unicas.copy())

    @staticmethod
    def actualizar_probabilidad(nodo: int, clase: str, probabilidad: float):
        Probabilidad.probabilidades[nodo][clase] = probabilidad

    @staticmethod
    def adicionar_en_tabla():
        query = "TRUNCATE TABLE algoritmos_arbol.clases_probabilidades"
        plpy.execute(query)

        values = []

        for nodo, clases in Probabilidad.probabilidades.items():
            for clase, probabilidad in clases.items():
                values.append(f"({nodo}, '{clase}', {probabilidad})")

        query = f"""
            INSERT INTO algoritmos_arbol.clases_probabilidades
            (nodo, clase, probabilidad) VALUES
            {",".join(values)}
        """
        plpy.execute(query)

    @staticmethod
    def eliminar_probabilidades_sueltas(id_nodos: List[int]):
        ids = list(Probabilidad.probabilidades.keys())
        for id in ids:
            if id not in id_nodos:
                del Probabilidad.probabilidades[id]

def es_numerico(columna: str) -> bool:
    query = f"""
    SELECT data_type AS tipo
    FROM information_schema.columns
    WHERE table_name = '{tabla}'
    AND table_schema = '{esquema}'
    AND column_name = '{columna}';
    """
    resultado = plpy.execute(query)
    tipo: str = resultado[0]["tipo"]
    return tipo in (
        "smallint",
        "integer",
        "bigint",
        "real",
        "double precision",
        "numeric",
    )


def entropia_clase(filtro: str = "true") -> Tuple[float, int]:
    query = f"SELECT count(*) as total FROM {esquema}.{tabla} WHERE {filtro}"

    resultado = plpy.execute(query)
    total_row = resultado[0]
    total_valor: int = total_row["total"]

    query = f"SELECT {atributo_objetivo}, count({atributo_objetivo}) as cuenta FROM {esquema}.{tabla} WHERE {filtro} GROUP BY {atributo_objetivo}"
    resultado = plpy.execute(query)

    entropia_valor = 0
    for fila in resultado:
        p_objetivo: float = fila["cuenta"] / total_valor
        entropia_valor -= p_objetivo * log2(p_objetivo)

    return entropia_valor, total_valor


def entropia_promedio(columna: str, filtro: str = "true") -> float:
    query = f"SELECT count(*) as total FROM {esquema}.{tabla} WHERE {filtro}"
    resultado = plpy.execute(query)
    total_row = resultado[0]
    total: int = total_row["total"]

    query = f"SELECT DISTINCT {columna} as valor FROM {esquema}.{tabla} WHERE {filtro}"
    resultado = plpy.execute(query)
    valores: List[str] = []
    for fila in resultado:
        valores.append(fila["valor"])

    entropia_total = 0
    for valor in valores:
        filtro_valor = f"({filtro}) AND {columna} = '{valor}'"
        entropia_valor, total_valor = entropia_clase(filtro_valor)

        entropia_total += (total_valor / total) * entropia_valor
    return entropia_total


class Nodo:
    id: int
    hijos: List["Nodo"]
    padre: Optional["Nodo"]
    tipo: Literal["RAIZ", "RAMA", "HOJA"]

    atributo: str
    atributo_valor: str

    clase: Optional[str]
    instancias_cantidad_correctas: int
    instancias_cantidad_incorrectas: int
    clase_total: int

    entropia: float

    atributos_disponibles: List[str]

    # Atributos globales
    contador_id: int = 0

    instancias_clasificadas_incorrectamente: int = 0
    instancias_clasificadas_correctamente: int = 0

    hojas: int = 0

    atributos_nominales_con_valores: Dict[str, List[str]] = {}

    def __init__(
        self,
        padre: Optional["Nodo"],
        tipo: Literal["RAIZ", "RAMA", "HOJA"],
        atributo: str,
        atributo_valor: str,
        atributos_disponibles: List[str],
    ) -> None:
        self.padre = padre
        self.tipo = tipo
        self.atributo = atributo
        self.atributo_valor = atributo_valor
        self.atributos_disponibles = atributos_disponibles
        self.hijos = []

        self.clase = None
        self.instancias_cantidad_correctas = 0
        self.instancias_cantidad_incorrectas = 0
        self.clase_total = 0

        # calculo entropia
        entropia, _ = entropia_clase(self.crear_filtro())
        self.entropia = entropia

        self.id = Nodo.contador_id
        Nodo.contador_id += 1

        if tipo == "HOJA":
            Nodo.hojas += 1

    def crear_filtro(self) -> str:
        filtro = "true"
        nodo = self
        while nodo != None:
            if nodo.tipo == "RAMA" or nodo.tipo == "HOJA":
                filtro += f" AND {nodo.atributo} = '{nodo.atributo_valor}'"

            nodo = nodo.padre

        return filtro

    def dar_clase(self):
        query = f"""
            SELECT {atributo_objetivo} AS clase, count(*) AS cantidad 
            FROM {esquema}.{tabla}
            WHERE {self.crear_filtro()}
            GROUP BY {atributo_objetivo} ORDER BY cantidad DESC;
        """
        resultado = plpy.execute(query)

        self.clase = resultado[0]["clase"]

        cantidades = [fila["cantidad"] for fila in resultado]

        total = sum(cantidades)

        self.clase_total = total

        self.instancias_cantidad_correctas = cantidades[0]
        Nodo.instancias_clasificadas_correctamente += self.instancias_cantidad_correctas

        self.instancias_cantidad_incorrectas = sum(cantidades[1:])
        Nodo.instancias_clasificadas_incorrectamente += (
            self.instancias_cantidad_incorrectas
        )

        for fila in resultado:
            Probabilidad.actualizar_probabilidad(
                self.id, fila["clase"], fila["cantidad"] / total
            )

    def crear_hijos(self):
        # elegir mejor atributo
        filtro = self.crear_filtro()
        ganancias: Dict[str, float] = {}
        for attr in self.atributos_disponibles:
            if es_numerico(attr):
                plpy.error(
                    f"La columna {attr} es de tipo númerico, "
                    + " no es compatible con este algoritmo"
                )
            ganancias[attr] = self.entropia - entropia_promedio(attr, filtro)

        mejor_ganancia: str = max(ganancias, key=lambda k: ganancias[k])

        query = f"""
            SELECT atributo_valor, count(*) AS contador
            FROM (SELECT {mejor_ganancia}::text AS atributo_valor, {atributo_objetivo}
                  FROM {esquema}.{tabla}
                  WHERE {filtro}
                  GROUP BY {mejor_ganancia}, {atributo_objetivo}) AS grupo
            GROUP BY atributo_valor ORDER BY contador;
        """
        resultado = plpy.execute(query)

        atributos_disponibles_hijo = self.atributos_disponibles[::]
        atributos_disponibles_hijo.remove(mejor_ganancia)

        # {atributo_valor: str, contador: int}
        for fila in resultado:
            if fila["contador"] <= 1 or len(atributos_disponibles_hijo) == 0:
                hijo = Nodo(
                    self,
                    "HOJA",
                    mejor_ganancia,
                    fila["atributo_valor"],
                    atributos_disponibles_hijo,
                )
                hijo.dar_clase()
            else:
                hijo = Nodo(
                    self,
                    "RAMA",
                    mejor_ganancia,
                    fila["atributo_valor"],
                    atributos_disponibles_hijo,
                )
                hijo.crear_hijos()

            self.hijos.append(hijo)

    def obtener_profundidad(self) -> int:
        profundidad = 0
        nodo = self.padre
        while nodo != None:
            profundidad += 1
            nodo = nodo.padre

        return profundidad

    def adicionar_nodos_en_tablas(self):
        id = self.id
        padre = self.padre.id if self.padre != None else -1
        tipo = self.tipo
        atributo = self.atributo
        atributo_valor = self.atributo_valor
        clase = "NULL" if self.clase is None else f"'{self.clase}'"
        instancias_cantidad_correctas = self.instancias_cantidad_correctas
        instancias_cantidad_incorrectas = self.instancias_cantidad_incorrectas

        plpy.execute(
            """
            INSERT INTO algoritmos_arbol.nodos
            (
                id,
                id_padre,
                tipo,
                atributo,
                comparador,
                atributo_valor,
                clase,
                instancias_cantidad_correctas,
                instancias_cantidad_incorrectas
            )
            VALUES (%s, %s, '%s', '%s', '%s', '%s', %s, %s, %s)
        """
            % (
                id,
                padre,
                tipo,
                atributo,
                "=",
                atributo_valor,
                clase,
                instancias_cantidad_correctas,
                instancias_cantidad_incorrectas,
            )
        )

        for hijo in self.hijos:
            hijo.adicionar_nodos_en_tablas()

    def obtener_id_nodos(self, lista: List[int]):
        lista.append(self.id)
        for hijo in self.hijos:
            hijo.obtener_id_nodos(lista)

    def cantidad_nodos(self):
        cantidad = 1
        for hijo in self.hijos:
            cantidad += hijo.cantidad_nodos()

        return cantidad

    def cantidad_hojas(self):
        if self.tipo == "HOJA":
            cantidad = 1
        else:
            cantidad = 0
        for hijo in self.hijos:
            cantidad += hijo.cantidad_hojas()

        return cantidad

    def obtener_cantidad_clase(self, cuenta: Dict[str, int]):
        if self.tipo == "HOJA" and self.clase is not None:
            cuenta[self.clase] += self.instancias_cantidad_correctas

        for hijo in self.hijos:
            hijo.obtener_cantidad_clase(cuenta)

    def obtener_clase_mas_probable(self):
        cantidad_clases: Dict[str, int] = defaultdict(int)
        self.obtener_cantidad_clase(cantidad_clases)
        return max(cantidad_clases, key=lambda k: cantidad_clases[k])


    def completar_atributos(self):
        if len(self.hijos) == 0:
            return

        for hijo in self.hijos:
            hijo.completar_atributos()

        atributo = self.hijos[0].atributo

        valores_hijos = [hijo.atributo_valor for hijo in self.hijos]
        valores_faltanes: List[str] = []

        if atributo not in Nodo.atributos_nominales_con_valores:
            return

        for valor in Nodo.atributos_nominales_con_valores[atributo]:
            if not valor in valores_hijos:
                valores_faltanes.append(valor)

        if len(valores_faltanes) == 0:
            return

        clase_mas_probable = self.obtener_clase_mas_probable()

        for valor in valores_faltanes:
            nodo = Nodo(self, "HOJA", atributo, valor, [])
            nodo.clase = clase_mas_probable
            Probabilidad.actualizar_probabilidad(nodo.id, nodo.clase, 0)
            self.hijos.append(nodo)

    def __repr__(self) -> str:
        return f"<Nodo tipo={self.tipo} atributo={self.atributo} atributo_valor={self.atributo_valor} entropia={self.entropia}>"


query = f"""
SELECT column_name as columna 
FROM information_schema.columns 
WHERE table_name = '{tabla}' 
AND table_schema = '{esquema}';
"""

resultado = plpy.execute(query)

atributos: List[str] = []

for fila in resultado:
    if fila["columna"] != atributo_objetivo:
        atributos.append(fila["columna"])

raiz = Nodo(None, "RAIZ", "-1", "-1", atributos)
raiz.crear_hijos()

# obtener atributos usados en cada nodo
atributos_nominales_usados: set[str] = set()


def agregar_atributos_usados(nodo: Nodo):
    if nodo.tipo != "RAIZ":
        atributos_nominales_usados.add(nodo.atributo)

    for hijo in nodo.hijos:
        agregar_atributos_usados(hijo)


agregar_atributos_usados(raiz)

for atributo in atributos_nominales_usados:
    query = f"""
        SELECT DISTINCT {atributo} as atributo
        FROM {esquema}.{tabla}
    """
    resultado = plpy.execute(query)
    valores = [fila["atributo"] for fila in resultado]
    Nodo.atributos_nominales_con_valores[atributo] = valores

raiz.completar_atributos()

raiz.adicionar_nodos_en_tablas()

ids = []
raiz.obtener_id_nodos(ids)
#Probabilidad.eliminar_probabilidades_sueltas(ids)
Probabilidad.adicionar_en_tabla()

def insertar_info(id, titulo, valor):
    plpy.execute(
        """
        INSERT INTO algoritmos_arbol.info (id, titulo, valor)
        VALUES ('%s', '%s', '%s')
    """
        % (id, titulo, valor)
    )


insertar_info("tipo", "Algoritmo Usado", "ID3")
insertar_info("tabla", "Tabla", tabla)
insertar_info("attr_objetivo", "Atributo Objetivo", atributo_objetivo)
insertar_info("esquema", "Esquema", esquema)
insertar_info("num_nodos", "Número de Nodos", raiz.cantidad_nodos())
insertar_info("num_hojas", "Número de Hojas", raiz.cantidad_hojas())
