use serde::Serialize;
use sqlx::prelude::FromRow;
use sqlx::Column;
use sqlx::Connection;
use sqlx::PgConnection;
use sqlx::Row;
use tauri::AppHandle;
use tokio::sync::Mutex;

use super::{emit_progress_event, get_connection, get_value_as_string, ConnectionManager};

#[derive(Serialize, FromRow)]
pub struct SummaryRow {
    cluster: i32,
    count: i32,
    percentage: f64,
}

#[derive(Serialize)]
pub struct RowResult {
    columns: Vec<String>,
    values: Vec<Vec<Option<String>>>,
}

#[derive(Serialize, FromRow)]
pub struct Point {
    x: i32,
    y: f64,
}

#[derive(Serialize)]
pub struct Info {
    name: String,
    value: String,
}

#[derive(Serialize)]
pub struct BestValue {
    best_k: i32,
    points: Vec<Point>,
}

#[tauri::command]
pub async fn load_table_for_kmeanspy(
    state: tauri::State<'_, Mutex<ConnectionManager>>,
    table: String,
    schema: String,
) -> Result<(), String> {
    let mut connection_manager = state.lock().await;
    let mut transaction = get_connection(&mut connection_manager)?
        .begin()
        .await
        .map_err(|e| format!("Error empezando una transacción: '{}'", e))?;

    let sql = "drop table if exists clustering.cl_data";
    sqlx::query(&sql)
        .execute(&mut *transaction)
        .await
        .map_err(|e| format!("Error eliminando la tabla clustering.cl_data: '{}'", e))?;

    let sql = format!(
        "create table clustering.cl_data as select * from {}.{}",
        schema, table
    );
    sqlx::query(&sql)
        .execute(&mut *transaction)
        .await
        .map_err(|e| {
            format!(
                "Error eliminando la tabla clustering.cl_data a partir de la tabla {}.{}: '{}'",
                schema, table, e
            )
        })?;

    transaction
        .commit()
        .await
        .map_err(|e| format!("Error tratando de completar la transacción: '{}'", e))?;

    Ok(())
}

async fn preprocess_table(app: &AppHandle, connection: &mut PgConnection) -> Result<(), String> {
    emit_progress_event(app, "Preprocesando la tabla...")?;
    sqlx::query("select clustering.preprocessing_py()")
        .execute(connection)
        .await
        .map_err(|e| format!("Error al tratar de preprocesar la tabla: '{}'", e))?;

    Ok(())
}

#[tauri::command]
pub async fn kmeans_elbow(
    state: tauri::State<'_, Mutex<ConnectionManager>>,
    app: tauri::AppHandle,
    max_clusters: i32,
    init_method: String,
    n_init: i32,
    algorithm: String,
    max_iter: i32,
    tolerance: f64,
    random_seed: i32,
) -> Result<Vec<Point>, String> {
    let mut connection_manager = state.lock().await;
    let mut transaction = get_connection(&mut connection_manager)?
        .begin()
        .await
        .map_err(|e| format!("Error empezando una transacción: '{}'", e))?;

    preprocess_table(&app, &mut *transaction).await?;

    emit_progress_event(&app, "Obteniendo valores de inercia...")?;

    sqlx::query(
        "SELECT clustering.elbow_py
            (
                n_max=>$1,
                init=>$2,
                n_init=>$3,
                algorithm=>$4,
                max_iter=>$5,
                tol=>$6,
                random_state=>$7
            ) AS result
        ",
    )
    .bind(max_clusters)
    .bind(init_method)
    .bind(n_init)
    .bind(algorithm)
    .bind(max_iter)
    .bind(tolerance)
    .bind(random_seed)
    .execute(&mut *transaction)
    .await
    .map_err(|e| {
        format!(
            "Error tratando de de obtener valores para graficar el codo: '{}'",
            e
        )
    })?;

    let result: Vec<Point> = sqlx::query_as(
        "
            SELECT
                k AS x,
                inertia AS y
            FROM
                clustering.cl_elbow
            ORDER BY k
        ",
    )
    .fetch_all(&mut *transaction)
    .await
    .map_err(|e| format!("Falló al tratar de consultar los puntos: '{}'", e))?;

    transaction
        .commit()
        .await
        .map_err(|e| format!("Error tratando de completar la transacción: '{}'", e))?;

    Ok(result)
}

#[tauri::command]
pub async fn kmeans_silhouette(
    state: tauri::State<'_, Mutex<ConnectionManager>>,
    app: tauri::AppHandle,
    max_clusters: i32,
    init_method: String,
    n_init: i32,
    algorithm: String,
    max_iter: i32,
    tolerance: f64,
    random_seed: i32,
) -> Result<BestValue, String> {
    let mut connection_manager = state.lock().await;
    let mut transaction = get_connection(&mut connection_manager)?
        .begin()
        .await
        .map_err(|e| format!("Error empezando una transacción: '{}'", e))?;

    preprocess_table(&app, &mut *transaction).await?;

    emit_progress_event(&app, "Obteniendo valores de coeficientes...")?;

    let result = sqlx::query(
        "SELECT 
            silhouette_py::Integer AS result
        FROM
            clustering.silhouette_py
                (
                    n_max=>$1,
                    init=>$2,
                    n_init=>$3,
                    algorithm=>$4,
                    max_iter=>$5,
                    tol=>$6,
                    random_state=>$7
                )
        ",
    )
    .bind(max_clusters)
    .bind(init_method)
    .bind(n_init)
    .bind(algorithm)
    .bind(max_iter)
    .bind(tolerance)
    .bind(random_seed)
    .fetch_one(&mut *transaction)
    .await
    .map_err(|e| format!("Error tratando de de obtener los coeficientes: '{}'", e))?;

    let best_point: i32 = result.get("result");

    let result: Vec<Point> = sqlx::query_as(
        "
            SELECT
                k AS x,
                silhouette_index AS y
            FROM
                clustering.cl_silhouette
            ORDER BY k
        ",
    )
    .fetch_all(&mut *transaction)
    .await
    .map_err(|e| format!("Falló al tratar de consultar los puntos: '{}'", e))?;

    transaction
        .commit()
        .await
        .map_err(|e| format!("Error tratando de completar la transacción: '{}'", e))?;

    Ok(BestValue {
        best_k: best_point,
        points: result,
    })
}

#[tauri::command]
pub async fn kmeans_main(
    state: tauri::State<'_, Mutex<ConnectionManager>>,
    app: tauri::AppHandle,
    n_clusters: i32,
    init_method: String,
    n_init: i32,
    algorithm: String,
    max_iter: i32,
    tolerance: f64,
    random_seed: i32,
) -> Result<Vec<Info>, String> {
    let mut connection_manager = state.lock().await;
    let mut transaction = get_connection(&mut connection_manager)?
        .begin()
        .await
        .map_err(|e| format!("Error empezando una transacción: '{}'", e))?;

    preprocess_table(&app, &mut *transaction).await?;

    emit_progress_event(&app, "Ejecutando el algoritmo K-Means...")?;
    let result = sqlx::query(
        "SELECT clustering.kmeans_py 
            (
                n_clusters=>$1,
                init=>$2,
                n_init=>$3,
                algorithm=>$4,
                max_iter=>$5,
                tol=>$6,
                random_state=>$7
            ) AS result
        ",
    )
    .bind(n_clusters)
    .bind(init_method)
    .bind(n_init)
    .bind(algorithm)
    .bind(max_iter)
    .bind(tolerance)
    .bind(random_seed)
    .fetch_one(&mut *transaction)
    .await
    .map_err(|e| format!("Error tratando de ejecutar K-means: '{}'", e))?;

    emit_progress_event(&app, "Obteniendo estadisticas...")?;
    let result: String = result.get("result");

    // Split values by ',' then by ':'
    let result: Vec<Vec<&str>> = result
        .split(',')
        .map(|s| s.split(':').map(|si| si.trim()).collect())
        .collect();

    // Map values into Info Struct
    let mut result: Vec<Info> = result
        .into_iter()
        .map(|sv| Info {
            name: sv[0].to_string(),
            value: sv[1].to_string(),
        })
        .collect();

    let count: i64 = sqlx::query("SELECT count(*) AS count FROM clustering.cl_data")
        .fetch_one(&mut *transaction)
        .await
        .map_err(|e| format!("Fallo al tratar de obtener la cantidad de filas: '{}'", e))?
        .get("count");

    result.push(Info {
        name: "row_count".into(),
        value: count.to_string(),
    });

    let count: i64 = sqlx::query(
        "
            SELECT count(*) AS count
            FROM information_schema.columns
            WHERE table_name = 'cl_data' 
            AND table_schema = 'clustering';
    ",
    )
    .fetch_one(&mut *transaction)
    .await
    .map_err(|e| {
        format!(
            "Fallo al tratar de obtener la cantidad de columnas: '{}'",
            e
        )
    })?
    .get("count");

    result.push(Info {
        name: "column_count".into(),
        value: count.to_string(),
    });

    transaction
        .commit()
        .await
        .map_err(|e| format!("Error tratando de completar la transacción: '{}'", e))?;

    Ok(result)
}

#[tauri::command]
pub async fn kmeans_summary(
    state: tauri::State<'_, Mutex<ConnectionManager>>,
    app: tauri::AppHandle,
) -> Result<Vec<SummaryRow>, String> {
    let mut connection_manager = state.lock().await;
    let connection = get_connection(&mut connection_manager)?;

    emit_progress_event(&app, "Generando resumen...")?;
    let schemas: Vec<SummaryRow> = sqlx::query_as(
        "
            SELECT
                cluster,
                count,
                percentage
            FROM
                clustering.summary_py()
        ",
    )
    .fetch_all(connection)
    .await
    .map_err(|e| format!("Falló al tratar de consultar el resumen: '{}'", e))?;

    Ok(schemas)
}

#[tauri::command]
pub async fn kmeans_centroids(
    state: tauri::State<'_, Mutex<ConnectionManager>>,
    app: tauri::AppHandle,
) -> Result<RowResult, String> {
    let mut connection_manager = state.lock().await;
    let mut transaction = get_connection(&mut connection_manager)?
        .begin()
        .await
        .map_err(|e| format!("Error empezando una transacción: '{}'", e))?;

    emit_progress_event(&app, "Obteniendo centroides...")?;

    sqlx::query("select * from clustering.centroids_py()")
        .persistent(false)
        .execute(&mut *transaction)
        .await
        .map_err(|e| format!("Error al tratar de generar los centroides: '{}'", e))?;

    let rows = sqlx::query("SELECT * FROM clustering.cl_centroids")
        .persistent(false)
        .fetch_all(&mut *transaction)
        .await
        .map_err(|e| format!("Falló al tratar de consultar el resumen: '{}'", e))?;

    if rows.is_empty() {
        return Ok(RowResult {
            columns: Vec::new(),
            values: Vec::new(),
        });
    }

    let columns: Vec<String> = rows[0]
        .columns()
        .iter()
        .map(|col| col.name().to_string())
        .collect();

    let mut all_values: Vec<Vec<Option<String>>> = Vec::new();

    for row in rows {
        let columns = row.columns();
        let mut values: Vec<Option<String>> = Vec::new();

        for column in columns {
            let column = column.name();
            let value = get_value_as_string(&row, column);
            values.push(value);
        }

        all_values.push(values);
    }

    transaction
        .commit()
        .await
        .map_err(|e| format!("Error tratando de completar la transacción: '{}'", e))?;

    Ok(RowResult {
        columns,
        values: all_values,
    })
}

#[tauri::command]
pub async fn kmeans_results(
    offset: i32,
    state: tauri::State<'_, Mutex<ConnectionManager>>,
    app: tauri::AppHandle,
) -> Result<RowResult, String> {
    let mut connection_manager = state.lock().await;
    let mut transaction = get_connection(&mut connection_manager)?
        .begin()
        .await
        .map_err(|e| format!("Error empezando una transacción: '{}'", e))?;

    let mut offset = offset;

    if offset < 0 {
        emit_progress_event(&app, "Generando resultados...")?;
        sqlx::query("select * from clustering.result_py()")
            .persistent(false)
            .execute(&mut *transaction)
            .await
            .map_err(|e| format!("Error al tratar de generar los resultados: '{}'", e))?;

        offset = 0;
    }

    let query = format!(
        "
            SELECT * 
            FROM clustering.cl_result
            OFFSET {} 
            LIMIT 10
        ",
        offset * 10
    );

    let rows = sqlx::query(&query)
        .persistent(false)
        .fetch_all(&mut *transaction)
        .await
        .map_err(|e| format!("Falló al tratar de consultar el resumen: '{}'", e))?;

    if rows.is_empty() {
        return Ok(RowResult {
            columns: Vec::new(),
            values: Vec::new(),
        });
    }

    let columns: Vec<String> = rows[0]
        .columns()
        .iter()
        .map(|col| col.name().to_string())
        .collect();

    let mut all_values: Vec<Vec<Option<String>>> = Vec::new();

    for row in rows {
        let columns = row.columns();
        let mut values: Vec<Option<String>> = Vec::new();

        for column in columns {
            let column = column.name();
            let value = get_value_as_string(&row, column);
            values.push(value);
        }

        all_values.push(values);
    }

    transaction
        .commit()
        .await
        .map_err(|e| format!("Error tratando de completar la transacción: '{}'", e))?;

    Ok(RowResult {
        columns,
        values: all_values,
    })
}
