use std::fs;

use csv::ReaderBuilder;
use encoding_rs::Encoding;
use serde::Deserialize;
use sqlx::Connection;
use tempfile::tempdir;
use tokio::sync::Mutex;

use super::{get_connection, ConnectionManager};

#[derive(Deserialize)]
pub enum EncodingFrontend {
    Utf8,
    Utf16,
    Latin1,
    Cp1252,
    Iso8859_2,
    Iso8859_5,
    Iso8859_15,
    Cp1251,
    Koi8R,
    ShiftJis,
    Gbk,
    Big5,
    EucJp,
    MacRoman,
}

impl EncodingFrontend {
    pub fn get_encoding(&self) -> &'static Encoding {
        match self {
            EncodingFrontend::Utf8 => encoding_rs::UTF_8,
            EncodingFrontend::Utf16 => encoding_rs::UTF_16LE,
            EncodingFrontend::Latin1 => encoding_rs::WINDOWS_1252,
            EncodingFrontend::Cp1252 => encoding_rs::WINDOWS_1252,
            EncodingFrontend::Iso8859_2 => encoding_rs::ISO_8859_2,
            EncodingFrontend::Iso8859_5 => encoding_rs::ISO_8859_5,
            EncodingFrontend::Iso8859_15 => encoding_rs::ISO_8859_15,
            EncodingFrontend::Cp1251 => encoding_rs::WINDOWS_1251,
            EncodingFrontend::Koi8R => encoding_rs::KOI8_R,
            EncodingFrontend::ShiftJis => encoding_rs::SHIFT_JIS,
            EncodingFrontend::Gbk => encoding_rs::GBK,
            EncodingFrontend::Big5 => encoding_rs::BIG5,
            EncodingFrontend::EucJp => encoding_rs::EUC_JP,
            EncodingFrontend::MacRoman => encoding_rs::MACINTOSH,
        }
    }

    pub fn get_encoding_sql(&self) -> &'static str {
        match self {
            EncodingFrontend::Utf8 => "UTF8",
            EncodingFrontend::Utf16 => "UTF16",
            EncodingFrontend::Latin1 => "LATIN1",
            EncodingFrontend::Cp1252 => "WIN1252",
            EncodingFrontend::Iso8859_2 => "LATIN2",
            EncodingFrontend::Iso8859_5 => "ISO_8859_5",
            EncodingFrontend::Iso8859_15 => "LATIN9",
            EncodingFrontend::Cp1251 => "WIN1251",
            EncodingFrontend::Koi8R => "KOI8R",
            EncodingFrontend::ShiftJis => "SJIS",
            EncodingFrontend::Gbk => "GBK",
            EncodingFrontend::Big5 => "BIG5",
            EncodingFrontend::EucJp => "EUC_JP",
            EncodingFrontend::MacRoman => "MAC",
        }
    }
}

#[derive(Deserialize)]
pub struct Header {
    name: String,
    datatype: String,
}

impl Header {
    pub fn as_string(&self) -> String {
        format!("{} {}", self.name, self.datatype)
    }
}

#[tauri::command]
pub fn get_csv_first_rows(
    file_path: String,
    delimiter: String,
    quote: String,
    encoding: EncodingFrontend,
) -> Result<Vec<Vec<String>>, String> {
    let raw_bytes =
        fs::read(file_path).map_err(|e| format!("Error leyendo el archivo: '{}'", e))?;

    let encoding = encoding.get_encoding();
    let (decoded, _, _) = encoding.decode(&raw_bytes);

    let delimiter = delimiter.as_bytes();

    if delimiter.len() > 1 {
        return Err(String::from("El delimitador debe ser un solo caracter"));
    }

    let delimiter = delimiter
        .first()
        .ok_or_else(|| String::from("Delimitador no valido"))?;

    let quote = quote.as_bytes();

    if quote.len() > 1 {
        return Err(String::from("El delimitador debe ser un solo caracter"));
    }
    let quote = quote
        .first()
        .ok_or_else(|| String::from("Comillas no validas"))?;

    let mut reader = ReaderBuilder::new()
        .delimiter(*delimiter)
        .quote(*quote)
        .from_reader(decoded.as_bytes());

    let headers = reader
        .headers()
        .map_err(|e| format!("Error obteniendo headers del archivo CSV: '{}'", e))?
        .iter()
        .map(|s| s.to_string())
        .collect::<Vec<String>>();

    let mut rows: Vec<Vec<String>> = vec![headers];
    let records_iterator = reader.records();

    for (i, record) in records_iterator.take(10).enumerate() {
        match record {
            Ok(row) => {
                rows.push(row.iter().map(|s| s.to_string()).collect());
            }
            Err(e) => {
                return Err(format!("Error leyendo la fila {}: '{}'", i + 1, e));
            }
        }
    }

    Ok(rows)
}

#[tauri::command]
pub async fn table_from_file(
    file_path: String,
    delimiter: String,
    quote: String,
    encoding: EncodingFrontend,
    null_string: String,
    table: String,
    schema: String,
    headers: Vec<Header>,
    state: tauri::State<'_, Mutex<ConnectionManager>>,
) -> Result<(), String> {
    let tmp_dir = tempdir().map_err(|e| format!("Error creando directorio temporal: {}", e))?;

    let destination = tmp_dir.path().join("archivo.csv");

    fs::copy(file_path, &destination)
        .map_err(|e| format!("Error copiando el archivo a la carpeta temporal: {}", e))?;

    let destination_str = destination.to_str().ok_or_else(|| {
        format!("Error obteniendo la ruta de destino, ¿Tiene caracteres Especiales?")
    })?;

    let mut connection_manager = state.lock().await;
    let mut transaction = get_connection(&mut connection_manager)?
        .begin()
        .await
        .map_err(|e| format!("Error empezando una transacción: '{}'", e))?;

    {
        let columns_names: Vec<String> = headers.iter().map(|c| c.as_string()).collect();
        let columns_names: String = columns_names.join(",");

        let sql = format!("CREATE TABLE {}.{} ({})", schema, table, columns_names);

        sqlx::query(&sql)
            .execute(&mut *transaction)
            .await
            .map_err(|e| format!("Fallo al tratar de crear la tabla: '{}'", e))?;
    }

    {
        let encoding = encoding.get_encoding_sql();

        let columns_names: Vec<String> = headers.iter().map(|c| c.name.clone()).collect();
        let columns_names: String = columns_names.join(",");

        let quote = if quote == "'" {
            String::from("'''")
        } else {
            quote
        };

        let sql = format!(
            "
                COPY {}.{} ({})
                FROM '{}'
                WITH (
                    FORMAT csv,
                    HEADER true,
                    DELIMITER '{}',
                    ENCODING '{}',
                    QUOTE '{}',
                    NULL '{}'
                )
            ",
            schema, table, columns_names, destination_str, delimiter, encoding, quote, null_string
        );

        sqlx::query(&sql)
            .execute(&mut *transaction)
            .await
            .map_err(|e| format!("Fallo al copiar el CSV: '{}'", e))?;
    }

    transaction
        .commit()
        .await
        .map_err(|e| format!("Error tratando de completar la transacción '{}'", e))?;
    Ok(())
}
