use serde::{Deserialize, Serialize};
use sqlx::postgres::PgConnectOptions;
use sqlx::postgres::PgRow;
use sqlx::prelude::FromRow;
use sqlx::types::BigDecimal;
use sqlx::Row;
use sqlx::{Connection, PgConnection};
use tauri::Emitter;
use tokio::sync::{Mutex, MutexGuard};
use tokio::time::{timeout, Duration};

pub mod clustering;
pub mod csv;
pub mod tree;

#[derive(Serialize, Deserialize)]
pub struct ConnectionDetails {
    name: String,
    host: String,
    port: u16,
    database: String,
    user: String,
    password: String,
}

#[derive(Serialize, FromRow)]
pub struct Distribution {
    row_num: i64,
    label: String,
    count: i64,
}

pub struct ConnectionManager {
    connection_details: Option<ConnectionDetails>,
    connection: Option<PgConnection>,
}

impl Default for ConnectionManager {
    fn default() -> Self {
        Self {
            connection_details: None,
            connection: None,
        }
    }
}

async fn create_connection(connection_details: &ConnectionDetails) -> Result<PgConnection, String> {
    let options = PgConnectOptions::new()
        .host(&connection_details.host)
        .port(connection_details.port)
        .username(&connection_details.user)
        .password(&connection_details.password)
        .database(&connection_details.database);

    let timeout_duration = Duration::from_secs(6);

    match timeout(timeout_duration, PgConnection::connect_with(&options)).await {
        Ok(connection_result) => connection_result
            .map_err(|e| format!("Falló la conexión al servidor de base de datos: '{}'", e)),
        Err(_) => {
            Err("Tiempo de espera agotado al intentar conectar a la base de datos".to_string())
        }
    }
}

#[tauri::command]
pub async fn database_connect(
    connection_details: ConnectionDetails,
    state: tauri::State<'_, Mutex<ConnectionManager>>,
) -> Result<i32, String> {
    let mut connection = create_connection(&connection_details).await?;

    let pid: i32 = sqlx::query_scalar("SELECT pg_backend_pid()")
        .fetch_one(&mut connection)
        .await
        .map_err(|e| format!("Error obteniendo PID de conexión: {}", e))?;

    let mut connection_manager = state.lock().await;

    connection_manager.connection = Some(connection);
    connection_manager.connection_details = Some(connection_details);

    Ok(pid)
}

#[tauri::command]
pub async fn cancel_query(connection_details: ConnectionDetails, pid: i32) -> Result<(), String> {
    let mut connection = create_connection(&connection_details).await?;

    let sql = format!("SELECT pg_cancel_backend({})", pid);

    sqlx::query(&sql)
        .execute(&mut connection)
        .await
        .map_err(|e| format!("Error Cancelando la conexión: {}", e))?;

    Ok(())
}

#[tauri::command]
pub async fn get_schemas(
    state: tauri::State<'_, Mutex<ConnectionManager>>,
) -> Result<Vec<String>, String> {
    let mut connection_manager = state.lock().await;
    let connection = get_connection(&mut connection_manager)?;

    let sql = "
        SELECT schema_name as schema
        FROM information_schema.schemata 
        WHERE schema_name NOT LIKE 'pg_%' 
        AND schema_name != 'information_schema';
    ";

    let query = sqlx::query(sql);

    let schemas = query
        .fetch_all(connection)
        .await
        .map_err(|e| format!("Falló al tratar de consultar los esquemas: '{}'", e))?;

    let schemas: Vec<String> = schemas.iter().map(|row| row.get("schema")).collect();

    Ok(schemas)
}

#[tauri::command]
pub async fn get_tables(
    schema: String,
    state: tauri::State<'_, Mutex<ConnectionManager>>,
) -> Result<Vec<String>, String> {
    let mut connection_manager = state.lock().await;
    let connection = get_connection(&mut connection_manager)?;

    let sql = r"
        SELECT table_name as table
        FROM information_schema.tables 
        WHERE table_schema = $1
        AND table_type = 'BASE TABLE';
    ";

    let query = sqlx::query(sql).bind(schema);

    let schemas = query
        .fetch_all(connection)
        .await
        .map_err(|e| format!("Falló al tratar de consultar las tablas: '{}'", e))?;

    let schemas: Vec<String> = schemas.iter().map(|row| row.get("table")).collect();

    Ok(schemas)
}

#[tauri::command]
pub async fn get_columns(
    schema: String,
    table: String,
    state: tauri::State<'_, Mutex<ConnectionManager>>,
) -> Result<Vec<String>, String> {
    let mut connection_manager = state.lock().await;
    let connection = get_connection(&mut connection_manager)?;

    let sql = r"
        SELECT column_name as column
        FROM information_schema.columns
        WHERE table_schema = $1
        AND table_name = $2;

    ";

    let query = sqlx::query(sql).bind(schema).bind(table);

    let columns = query
        .fetch_all(connection)
        .await
        .map_err(|e| format!("Falló al tratar de consultar las columnas: '{}'", e))?;

    let columns: Vec<String> = columns.iter().map(|row| row.get("column")).collect();

    Ok(columns)
}

#[tauri::command]
pub async fn get_table_row_count(
    schema: String,
    table: String,
    state: tauri::State<'_, Mutex<ConnectionManager>>,
) -> Result<i64, String> {
    let mut connection_manager = state.lock().await;
    let connection = get_connection(&mut connection_manager)?;

    let sql = format!(
        r"
        SELECT count(*) as count
        FROM {}.{}
    ",
        schema, table
    );

    let query = sqlx::query(&sql).bind(schema);

    let row = query
        .fetch_one(connection)
        .await
        .map_err(|e| format!("Falló al tratar de contar las filas: '{}'", e))?;

    let count: i64 = row.get("count");

    Ok(count)
}

#[tauri::command]
pub async fn get_column_data_distribution(
    schema: String,
    table: String,
    column: String,
    state: tauri::State<'_, Mutex<ConnectionManager>>,
) -> Result<Vec<Distribution>, String> {
    let mut connection_manager = state.lock().await;
    let connection = get_connection(&mut connection_manager)?;

    let sql = format!(
        "
            SELECT 
                row_number() OVER () AS row_num,
                \"{}\" AS label,
                count(*) AS count
            FROM {}.{}
            GROUP BY \"{}\"
    ",
        column, schema, table, column
    );

    let columns = sqlx::query(&sql)
        .fetch_all(connection)
        .await
        .map_err(|e| format!("Falló al tratar de consultar la distribución: '{}'", e))?;

    let columns: Vec<Distribution> = columns
        .iter()
        .map(|row| Distribution {
            row_num: row.get("row_num"),
            label: get_value_as_string(row, "label").unwrap_or("NULL".to_string()),
            count: row.get("count"),
        })
        .collect();

    Ok(columns)
}

#[tauri::command]
pub async fn cut_table_columns(
    source_schema: String,
    source_table: String,
    new_schema: String,
    new_table: String,
    overwrite: bool,
    columns: Vec<String>,
    state: tauri::State<'_, Mutex<ConnectionManager>>,
    app: tauri::AppHandle,
) -> Result<(), String> {
    let mut connection_manager = state.lock().await;

    let mut transaction = get_connection(&mut connection_manager)?
        .begin()
        .await
        .map_err(|e| format!("Error empezando una transacción: '{}'", e))?;

    if overwrite {
        if source_schema == new_schema && source_table == new_table {
            return Err("No puede sobrescribir la tabla de origen".into());
        }

        let sql = format!("DROP TABLE IF EXISTS {}.{}", new_schema, new_table);
        sqlx::query(&sql)
            .execute(&mut *transaction)
            .await
            .map_err(|e| format!("Falló al tratar de eliminar los datos de la tabla: '{}'", e))?;
    }

    emit_progress_event(&app, "Copiando tabla Tabla...")?;

    let columns: Vec<String> = columns.iter().map(|c| format!("\"{}\"", c)).collect();
    let columns = columns.join(", ");

    let sql = format!(
        "
            CREATE TABLE {}.{} AS
            SELECT {}
            FROM {}.{};
    ",
        new_schema, new_table, columns, source_schema, source_table
    );
    sqlx::query(&sql)
        .execute(&mut *transaction)
        .await
        .map_err(|e| format!("Falló al tratar de copiar la tabla: '{}'", e))?;

    transaction
        .commit()
        .await
        .map_err(|e| format!("Error tratando de completar la transacción '{}'", e))?;
    Ok(())
}

#[tauri::command]
pub async fn divide_table_manual(
    source_schema: String,
    source_table: String,
    training_schema: String,
    training_table: String,
    test_schema: String,
    test_table: String,
    overwrite: bool,
    percentage: i32,
    state: tauri::State<'_, Mutex<ConnectionManager>>,
    app: tauri::AppHandle,
) -> Result<String, String> {
    let mut connection_manager = state.lock().await;
    let connection = get_connection(&mut connection_manager)?;

    let sql = r"
        SELECT *
        FROM algoritmos_arbol.dividir_tabla_con_nombre($1, $2, $3, $4, $5, $6, $7, $8)
    ";

    let query = sqlx::query(&sql)
        .bind(source_table)
        .bind(source_schema)
        .bind(percentage)
        .bind(training_schema)
        .bind(training_table)
        .bind(test_schema)
        .bind(test_table)
        .bind(overwrite);

    emit_progress_event(&app, "Dividiendo Tabla...")?;

    query
        .fetch_one(connection)
        .await
        .map_err(|e| format!("Falló al tratar de dividir la tabla: '{}'", e))?;

    Ok("ok".into())
}

#[tauri::command]
pub async fn database_disconnect(
    state: tauri::State<'_, Mutex<ConnectionManager>>,
) -> Result<String, String> {
    let mut connection_manager = state.lock().await;

    connection_manager.connection = None;
    connection_manager.connection_details = None;

    Ok("Se desconectó de manera exitosa".into())
}

#[tauri::command]
pub async fn test_connection(
    state: tauri::State<'_, Mutex<ConnectionManager>>,
) -> Result<String, String> {
    let mut connection_manager = state.lock().await;
    let connection = get_connection(&mut connection_manager)?;

    let res = sqlx::query("SELECT version() as version")
        .fetch_one(connection)
        .await
        .expect("error en la conexión");

    let version = res.get("version");

    Ok(version)
}

fn get_connection<'a>(
    connection_manager: &'a mut MutexGuard<ConnectionManager>,
) -> Result<&'a mut sqlx::PgConnection, String> {
    connection_manager
        .connection
        .as_mut()
        .ok_or("No hay una conexión Activa".to_string())
}

#[derive(Clone, serde::Serialize)]
struct EventPayload {
    message: String,
}

fn emit_progress_event(app: &tauri::AppHandle, message: &str) -> Result<(), String> {
    app.emit(
        "progress",
        EventPayload {
            message: message.into(),
        },
    )
    .map_err(|e| format!("Error al enviar evento: {}", e))
}

fn get_value_as_string(row: &PgRow, col_name: &str) -> Option<String> {
    if let Ok(val) = row.try_get::<Option<String>, &str>(col_name) {
        return val;
    }
    if let Ok(val) = row.try_get::<Option<i64>, &str>(col_name) {
        return val.map(|v| v.to_string());
    }
    if let Ok(val) = row.try_get::<Option<i32>, &str>(col_name) {
        return val.map(|v| v.to_string());
    }
    if let Ok(val) = row.try_get::<Option<i16>, &str>(col_name) {
        return val.map(|v| v.to_string());
    }
    if let Ok(val) = row.try_get::<Option<i8>, &str>(col_name) {
        return val.map(|v| v.to_string());
    }
    if let Ok(val) = row.try_get::<Option<f64>, &str>(col_name) {
        return val.map(|v| v.to_string());
    }
    if let Ok(val) = row.try_get::<Option<f32>, &str>(col_name) {
        return val.map(|v| v.to_string());
    }
    if let Ok(val) = row.try_get::<Option<bool>, &str>(col_name) {
        return val.map(|v| v.to_string());
    }
    if let Ok(val) = row.try_get::<Option<BigDecimal>, &str>(col_name) {
        return val.map(|v| v.to_string());
    }
    None
}
