use super::{emit_progress_event, get_connection, ConnectionManager};
use sqlx::Connection;
use sqlx::PgConnection;
use sqlx::Row;
use tokio::sync::Mutex;

async fn clear_previous_evaluation(
    connection: &mut PgConnection,
    app: &tauri::AppHandle,
) -> Result<(), String> {
    emit_progress_event(app, "Eliminando texto de ejecución anterior...")?;

    sqlx::query("DELETE FROM algoritmos_arbol.textos_evaluacion")
        .execute(connection)
        .await
        .map_err(|e| {
            format!(
                "Fallo al tratar de eliminar textos de ejecución anterior: '{}'",
                e
            )
        })?;

    Ok(())
}

async fn create_j48_tree(
    table: &str,
    atribute: &str,
    schema: &str,
    minimal_objects: i64,
    confidence_factor: f64,
    connection: &mut PgConnection,
    app: &tauri::AppHandle,
) -> Result<(), String> {
    emit_progress_event(&app, "Creando Arbol J48...")?;

    let sql = format!(
        r"
            SELECT * FROM
            algoritmos_arbol.crear_arbol_c45(
                '{}', '{}', '{}', {}, {}
            )
        ",
        table, atribute, schema, minimal_objects, confidence_factor
    );

    sqlx::query(&sql)
        .execute(connection)
        .await
        .map_err(|e| format!("Fallo al tratar de crear el arbol J48: '{}'", e))?;

    Ok(())
}

async fn create_id3_tree(
    table: &str,
    atribute: &str,
    schema: &str,
    connection: &mut PgConnection,
    app: &tauri::AppHandle,
) -> Result<(), String> {
    emit_progress_event(&app, "Creando Arbol ID3...")?;

    let sql = format!(
        r"
            SELECT * FROM
            algoritmos_arbol.crear_arbol_id3(
                '{}', '{}', '{}'
            )
        ",
        table, atribute, schema
    );

    sqlx::query(&sql)
        .execute(connection)
        .await
        .map_err(|e| format!("Fallo al tratar de crear el arbol ID3: '{}'", e))?;

    Ok(())
}

async fn evaluate_table(
    table: &str,
    schema: &str,
    connection: &mut PgConnection,
    app: &tauri::AppHandle,
) -> Result<(), String> {
    emit_progress_event(&app, "Evaluando Arbol...")?;

    let sql = format!(
        r"
            SELECT * FROM
            algoritmos_arbol.evaluar_tabla('{}', '{}');
            ",
        table, schema
    );

    sqlx::query(&sql)
        .execute(connection)
        .await
        .map_err(|e| format!("Fallo al tratar de evaluar el arbol: '{}'", e))?;

    Ok(())
}

async fn generate_evaluation_text(
    connection: &mut PgConnection,
    app: &tauri::AppHandle,
) -> Result<(), String> {
    emit_progress_event(&app, "Generando textos de evaluación...")?;

    sqlx::query("SELECT * FROM algoritmos_arbol.mostrar()")
        .execute(connection)
        .await
        .map_err(|e| format!("Fallo al tratar de crear los textos de evaluación: '{}'", e))?;

    Ok(())
}

async fn get_result_text(
    connection: &mut PgConnection,
    app: &tauri::AppHandle,
) -> Result<String, String> {
    emit_progress_event(&app, "Obteniendo textos de evaluación...")?;

    let result = sqlx::query("SELECT valor FROM algoritmos_arbol.textos_evaluacion")
        .fetch_all(connection)
        .await
        .map_err(|e| format!("Fallo al tratar Obtener el resultado: '{}'", e))?;

    let mut text = String::new();

    for row in result.iter() {
        let value: String = row.get("valor");
        text.push_str(&value);
    }

    Ok(text)
}

#[tauri::command]
pub async fn evaluation_training_table(
    algorithm: String,
    table: String,
    atribute: String,
    schema: String,
    minimal_objects: i64,
    confidence_factor: f64,
    state: tauri::State<'_, Mutex<ConnectionManager>>,
    app: tauri::AppHandle,
) -> Result<String, String> {
    let mut connection_manager = state.lock().await;
    let mut transaction = get_connection(&mut connection_manager)?
        .begin()
        .await
        .map_err(|e| format!("Error empezando una transacción: '{}'", e))?;

    clear_previous_evaluation(&mut *transaction, &app).await?;

    if algorithm == "ID3" {
        create_id3_tree(&table, &atribute, &schema, &mut *transaction, &app).await?;
    } else {
        create_j48_tree(
            &table,
            &atribute,
            &schema,
            minimal_objects,
            confidence_factor,
            &mut *transaction,
            &app,
        )
        .await?;
    }

    evaluate_table(&table, &schema, &mut *transaction, &app).await?;

    generate_evaluation_text(&mut *transaction, &app).await?;

    let result_text = get_result_text(&mut *transaction, &app).await?;

    transaction
        .commit()
        .await
        .map_err(|e| format!("Error tratando de completar la transacción '{}'", e))?;

    Ok(result_text)
}

#[tauri::command]
pub async fn evaluation_specific_table(
    algorithm: String,
    table: String,
    atribute: String,
    schema: String,
    minimal_objects: i64,
    confidence_factor: f64,
    evaluation_table: String,
    evaluation_schema: String,
    state: tauri::State<'_, Mutex<ConnectionManager>>,
    app: tauri::AppHandle,
) -> Result<String, String> {
    let mut connection_manager = state.lock().await;
    let mut transaction = get_connection(&mut connection_manager)?
        .begin()
        .await
        .map_err(|e| format!("Error empezando una transacción: '{}'", e))?;

    clear_previous_evaluation(&mut *transaction, &app).await?;

    if algorithm == "ID3" {
        create_id3_tree(&table, &atribute, &schema, &mut *transaction, &app).await?;
    } else {
        create_j48_tree(
            &table,
            &atribute,
            &schema,
            minimal_objects,
            confidence_factor,
            &mut *transaction,
            &app,
        )
        .await?;
    }

    evaluate_table(
        &evaluation_table,
        &evaluation_schema,
        &mut *transaction,
        &app,
    )
    .await?;

    generate_evaluation_text(&mut *transaction, &app).await?;

    let result_text = get_result_text(&mut *transaction, &app).await?;

    transaction
        .commit()
        .await
        .map_err(|e| format!("Error tratando de completar la transacción '{}'", e))?;

    Ok(result_text)
}

#[tauri::command]
pub async fn predict_table_values(
    algorithm: String,
    table: String,
    atribute: String,
    schema: String,
    minimal_objects: i64,
    confidence_factor: f64,
    evaluation_table: String,
    evaluation_schema: String,
    instances_amount: i64,
    state: tauri::State<'_, Mutex<ConnectionManager>>,
    app: tauri::AppHandle,
) -> Result<String, String> {
    let mut connection_manager = state.lock().await;
    let mut transaction = get_connection(&mut connection_manager)?
        .begin()
        .await
        .map_err(|e| format!("Error empezando una transacción: '{}'", e))?;

    clear_previous_evaluation(&mut *transaction, &app).await?;

    if algorithm == "ID3" {
        create_id3_tree(&table, &atribute, &schema, &mut *transaction, &app).await?;
    } else {
        create_j48_tree(
            &table,
            &atribute,
            &schema,
            minimal_objects,
            confidence_factor,
            &mut *transaction,
            &app,
        )
        .await?;
    }

    emit_progress_event(&app, "Creando textos de evaluación para el arbol...")?;

    sqlx::query(
        "
          SELECT * FROM algoritmos_arbol.mostrar_arbol();
        ",
    )
    .execute(&mut *transaction)
    .await
    .map_err(|e| format!("Fallo al tratar de crear los textos de evaluación: '{}'", e))?;

    sqlx::query(
        "
          SELECT * FROM algoritmos_arbol.mostrar_info_arbol();
        ",
    )
    .execute(&mut *transaction)
    .await
    .map_err(|e| format!("Fallo al tratar de crear los textos de evaluación: '{}'", e))?;

    emit_progress_event(&app, "Prediciendo valores de la tabla...")?;

    let sql = format!(
        "SELECT * FROM algoritmos_arbol.predecir_tabla('{}', '{}', {})",
        evaluation_table, evaluation_schema, instances_amount
    );

    sqlx::query(&sql)
        .execute(&mut *transaction)
        .await
        .map_err(|e| format!("Fallo al tratar de predecir la tabla: '{}'", e))?;

    let result_text = get_result_text(&mut *transaction, &app).await?;

    transaction
        .commit()
        .await
        .map_err(|e| format!("Error tratando de completar la transacción '{}'", e))?;

    Ok(result_text)
}

#[tauri::command]
pub async fn evaluation_crossfold_validation(
    algorithm: String,
    table: String,
    atribute: String,
    schema: String,
    minimal_objects: i64,
    confidence_factor: f64,
    fold_number: i64,
    state: tauri::State<'_, Mutex<ConnectionManager>>,
    app: tauri::AppHandle,
) -> Result<String, String> {
    let mut connection_manager = state.lock().await;
    let mut transaction = get_connection(&mut connection_manager)?
        .begin()
        .await
        .map_err(|e| format!("Error empezando una transacción: '{}'", e))?;

    clear_previous_evaluation(&mut *transaction, &app).await?;

    {
        let message = format!("Evaluando Arbol {} con Crossfold Validation...", algorithm);
        emit_progress_event(&app, &message)?;

        let parameters = if algorithm == "ID3" {
            String::from("")
        } else {
            format!(", {}, {}", minimal_objects, confidence_factor)
        };

        let algorithm = if algorithm == "ID3" {
            String::from("id3")
        } else {
            String::from("c45")
        };

        let sql = format!(
            r"
                SELECT * FROM algoritmos_arbol.cross_validation('{}', '{}', '{}', '{}', '{}', {});
            ",
            table, atribute, schema, algorithm, parameters, fold_number
        );

        sqlx::query(&sql)
            .execute(&mut *transaction)
            .await
            .map_err(|e| {
                format!(
                    "Fallo al tratar de evaluar el arbol {} con Crossfold Validation: '{}'",
                    algorithm, e
                )
            })?;
    }

    generate_evaluation_text(&mut *transaction, &app).await?;

    let result_text = get_result_text(&mut *transaction, &app).await?;

    transaction
        .commit()
        .await
        .map_err(|e| format!("Error tratando de completar la transacción '{}'", e))?;

    Ok(result_text)
}

#[tauri::command]
pub async fn evaluation_percentage_split(
    algorithm: String,
    table: String,
    atribute: String,
    schema: String,
    minimal_objects: i64,
    confidence_factor: f64,
    percentage: i64,
    state: tauri::State<'_, Mutex<ConnectionManager>>,
    app: tauri::AppHandle,
) -> Result<String, String> {
    let mut connection_manager = state.lock().await;
    let mut transaction = get_connection(&mut connection_manager)?
        .begin()
        .await
        .map_err(|e| format!("Error empezando una transacción: '{}'", e))?;

    clear_previous_evaluation(&mut *transaction, &app).await?;

    let training = format!("{}_{}_e{}", schema, table, percentage);
    let test = format!("{}_{}_p{}", schema, table, 100 - percentage);

    {
        emit_progress_event(&app, "Dividiendo Tabla...")?;

        let sql = format!(
            r"
                SELECT * FROM algoritmos_arbol.dividir_tabla(
                    '{}', '{}', '{}'
                );
            ",
            table, schema, percentage
        );

        sqlx::query(&sql)
            .execute(&mut *transaction)
            .await
            .map_err(|e| format!("Fallo al tratar de dividir las tablas: '{}'", e))?;
    }

    if algorithm == "ID3" {
        create_id3_tree(
            &training,
            &atribute,
            "algoritmos_arbol_datos",
            &mut *transaction,
            &app,
        )
        .await?;
    } else {
        create_j48_tree(
            &training,
            &atribute,
            "algoritmos_arbol_datos",
            minimal_objects,
            confidence_factor,
            &mut *transaction,
            &app,
        )
        .await?;
    }

    evaluate_table(&test, "algoritmos_arbol_datos", &mut *transaction, &app).await?;

    generate_evaluation_text(&mut *transaction, &app).await?;

    let result_text = get_result_text(&mut *transaction, &app).await?;

    transaction
        .commit()
        .await
        .map_err(|e| format!("Error tratando de completar la transacción '{}'", e))?;

    Ok(result_text)
}
